/*
 * Copyright (C) 2012, Nils Asmussen <nils@os.inf.tu-dresden.de>
 * Economic rights: Technische Universitaet Dresden (Germany)
 *
 * This file is part of NRE (NOVA runtime environment).
 *
 * NRE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * NRE is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License version 2 for more details.
 */

#include <ipc/Service.h>
#include <stream/Serial.h>
#include <String.h>

#include "VirtualMemory.h"
#include "Hypervisor.h"
#include "Log.h"

using namespace nre;

BufferedLog BufferedLog::_inst INIT_PRIO_SERIAL;
Log Log::_inst INIT_PRIO_SERIAL;
Service *Log::_srv;
const char *Log::_colors[] = {
    "31", "32", "33", "34", "35", "36"
};

BufferedLog::BufferedLog() : BaseSerial(), _bufpos(0), _buf() {
    BaseSerial::_inst = this;
}

Ports::port_t Log::get_com1_base() {
    uintptr_t virt = VirtualMemory::alloc(ExecEnv::PAGE_SIZE);
    Hypervisor::map_mem(0x0, virt, ExecEnv::PAGE_SIZE);
    Ports::port_t base = *reinterpret_cast<uint16_t*>(virt + BDA_COM_PORTS_OFF);
    Hypervisor::unmap_mem(virt, ExecEnv::PAGE_SIZE);
    VirtualMemory::free(virt, ExecEnv::PAGE_SIZE);
    return base;
}

Log::Log() : BaseSerial(), _ports(get_com1_base(), 6), _sm(1), _ready(true) {
    _ports.out<uint8_t>(0x80, LCR);          // Enable DLAB (set baud rate divisor)
    _ports.out<uint8_t>(0x01, DLR_LO);       // Set divisor to 1 (lo byte) 115200 baud
    _ports.out<uint8_t>(0x00, DLR_HI);       //                  (hi byte)
    _ports.out<uint8_t>(0x03, LCR);          // 8 bits, no parity, one stop bit
    _ports.out<uint8_t>(0, IER);             // disable interrupts
    _ports.out<uint8_t>(7, FCR);
    _ports.out<uint8_t>(3, MCR);
}

void Log::start() {
    _srv = new LogService("log");
    _srv->start();
}

void Log::write(const char *name, uint sessid, const char *line, size_t len) {
    ScopedLock<UserSm> guard(&_sm);
    *this << "\e[0;" << _colors[sessid % ARRAY_SIZE(_colors)] << "m[" << fmt(name, 8, 8) << "] ";
    for(size_t i = 0; i < len; ++i) {
        char c = line[i];
        if(c != '\n')
            *this << c;
    }
    *this << "\e[0m\n";
}

void Log::LogService::portal(LogServiceSession *sess) {
    UtcbFrameRef uf;
    try {
        String line;
        uf >> line;
        uf.finish_input();

        Log::get().write(sess->name().str(), sess->id() + 1, line.str(), line.length());
        uf << E_SUCCESS;
    }
    catch(const Exception &e) {
        uf.clear();
        uf << e;
    }
}
